from sage.all import Matrix
from sage.all import Rational
from sage.all import Integer
from sage.all import vector
from sage.all import Graphics
from sage.all import text3d
from sage.all import Polyhedron
from sage.all import gcd
from sage.all import floor
from sage.all import Infinity
from sage.all import n
from sage.all import text
from sage.all import text3d
from sage.all import RR
from sage.all import VectorSpace
from sage.all import QQ
from sage.all import lcm

from sage.misc.misc import SAGE_TMP



from __future__ import print_function

import webbrowser
import time

from tnorm.x3d_to_html import make_x3d_html
from tnorm.sage_types import *
from tnorm.hasse import get_hasse
from tnorm.utilities import *


class NormBall():
    def __init__(self,vertices, polyhedron):
        self.vertices = vertices
        self.polyhedron = polyhedron
        self.dimension = polyhedron.dim()
        self.is_full_dimensional = self.polyhedron.is_full_dimensional()
        self.is_compact = self.polyhedron.is_compact()
        self.num_vertices = len(self.polyhedron.faces(0))
        self.num_faces = len(self.polyhedron.faces(self.polyhedron.dim()-1))
        self.face_list = self.polyhedron.faces(self.polyhedron.dim()-1)
        self._face_dict = {}

    ### sage does not index lines, rays, and vertices independently. We would like our indices to agree with those 
    ### of the sage Polyhedron, with lines and vertices indexed separately, so instead of using sages indices 
    ### via vertex.index(), we take the index of a vertex (or line) to be the index of the vertex (or line) in the tuple returned
    ### by self.polyhedron.vertices() (or self.polyhedron.lines()). This tuple is cached by sage, so the the order of the vertices in this 
    ### tuple should never change. The below function fetches this index.
    def index_of_poly_vert(self, v, is_line=False):   ## or line
        if is_line:
            verts = self.polyhedron.lines()
        else:
            verts = self.polyhedron.vertices()
        for i in range(len(verts)):
            if v == verts[i]:
                return i

    def vertex(self, index):
        return self.vertices[index]

    def show_schlegel_diagram(self):
        G = Graphics()
        sp = self.polyhedron.schlegel_projection()
        p = sp.plot()
        G += p
        G.show(viewer='threejs', online=True)

    @cached_property
    def get_hasse_diagram(self):
        G = get_hasse(self.polyhedron)
        return G

    def show_hasse_diagram(self):
        G = self.get_hasse_diagram
        G.show()

    def get_hasse_graph(self):
        return self.polyhedron.face_lattice().hasse_diagram()


    def _plot2d(self, dual=False):
        labels = Graphics()
        for i in range(self.num_vertices):
            v=self.vertices[i]
            if not dual:
                labels+=text('S_{},{}'.format(v.genus,v.num_boundary_comps), tuple(v.coords[j]*1.15 for j in range(len(v.coords))), color='black', dpi=200, fontsize='large', horizontal_alignment='center')
            else:
                labels+=text('{}'.format(v.index), tuple(v.coords[j]*1.15 for j in range(len(v.coords))), color='black', dpi=200, fontsize='large', horizontal_alignment='center')
        plt = self.polyhedron.plot(dpi=200) + labels
        return plt

    def plot(self, viewer='x3d', online=False, opacity=1):
        dual = True if isinstance(self, DualNormBall) else False
        P = self.polyhedron
        if not P.is_compact():
            P = Polyhedron(P.vertices_list()) # we can't plot an infinite polyhedron, so just plot the finite part.
        if P.dim() in [1,2]:
            plt = self._plot2d(dual)
            plt.show()
        elif P.dim() in [3,4]:
            if viewer == 'x3d':
                with make_temp_directory() as temp_dir:
                    filename = make_x3d_html(self, online=online, directory=temp_dir, dual=dual)
                    webbrowser.open('file://'+filename)
                    time.sleep(5)
            else:
                labels=Graphics()
                for i in range(self.num_vertices):
                    v=self.vertices[i]
                    if not dual:
                        labels+=text3d('S_{},{}'.format(v.genus,v.num_boundary_comps),(v.coords[0]*1.15,v.coords[1]*1.15,v.coords[2]*1.15),color='black',dpi=200,fontsize='large',horizontal_alignment='center')
                    else:
                        labels+=text3d('{}'.format(v.index),(v.coords[0]*1.15,v.coords[1]*1.15,v.coords[2]*1.15),color='black',dpi=200,fontsize='large',horizontal_alignment='center')
                p = P.plot(opacity=opacity, edge_thickness=2)
                G = p+labels
                if with_dual:
                    pd = self.dual().plot(opacity=opacity, edge_thickness=2, color='blue')
                    G += pd
                G.show(viewer=viewer,online=online,axes=True)
        else:
            print('Can\'t plot polyhedron of dimension greater than 4.')

class TNormBall(NormBall):
    def __init__(self, vertices, rays, polyhedron):
        NormBall.__init__(self, vertices, polyhedron)
        self.rays = rays
        if not self.is_compact:
            self.polyhedron_mod_rays = Polyhedron(polyhedron.vertices_list())
        self.num_rays = len(self.polyhedron.rays())+2*len(self.polyhedron.lines())
        self.all_vertices_admissible = True
        for v in self.vertices:
             if not v.is_admissible:
                self.all_vertices_admissible = False
        
        for dim in range(1,self.polyhedron.dim()):
            self._face_dict[dim] = []
            for i in range(len(self.polyhedron.faces(dim))):
                face = self.polyhedron.faces(dim)[i]
                vertices = [self.vertex(self.index_of_poly_vert(v)) for v in face.vertices()]
                rays = [self.ray(2*self.index_of_poly_vert(l, True)+j) for l in face.lines() for j in [0,1]]
                self._face_dict[dim].append(Facet(i, dim, vertices, rays))



    def facets(self, dim):
        if dim == 0:
            return self.vertices
        else:
            return self._face_dict[dim]

    def facet(self, dim, index):
        if dim == 0:
            return self.vertices[index]
        else:
            return self._face_dict[dim][index]

    def num_facets(dim):
        return len(self.facets(dim))


    def ray(self, index):
        return self.rays[index]

    def project_to_norm(self):
        """ if the norm ball is not compact, project the infinite polyhedron to
            a subspace in which the Thurston norm is a true norm. While B.polyhedron_mod_rays
            is such a projection of P, it still lives in the same ambient space as P. This 
            function returns a polyhedron that is full dimensional in its ambient space (i.e.,
            the space is projected as well as the polyhedron). If B is compact, this just
            returns P (i.e., it does nothing). Also, if the dimension of P is < 4, this just returns
            B.polyhedron_mod_rays. This is becuase 
        """
        if B.polyhedron.dim() < 4:
            return self.polyhedron_mod_rays
        else:
            P = self.polyhedron_mod_rays
            subspace = P.change_ring(RR).ambient_space().subspace(P.vertices_list())
            basis = subspace.basis()
            projected_verts = [[vector(v).dot_product(b.normalized()) for b in basis] for v in P.vertices_list()]
            return Polyhedron(projected_verts)

#        zeros = vector([0 for i in range(n)])
#        M = Matrix(subspace.basis() + [zeros for i in range(n-k)])
#        perp = M.solve_right(zeros)




class DualNormBall(NormBall):
    def __init__(self, vertices, polyhedron):
        NormBall.__init__(self, vertices, polyhedron)
        assert self.is_compact == True  # the dual norm ball should always be compact


class Vertex():
    def __init__(self, index, coords):
        self.index = index
        self.coords = coords


class AdornedVertex(Vertex):
    def __init__(self, index, surface_index, coords, num_boundary_comps, euler, boundary_slopes, is_ray, is_admissible, has_rep):
        Vertex.__init__(self, index, coords)
        self.surface_index = surface_index
        self.has_surface_rep = has_rep
        self.num_boundary_comps = num_boundary_comps 
        self.euler_char = euler
        self.boundary_slopes = boundary_slopes
        self._is_ray = is_ray
        if self.has_surface_rep:
            self.frac = (-1./self.euler_char) if self.euler_char != 0 else +Infinity
            self.genus = (2- self.euler_char - self.num_boundary_comps)/2
        else:
            self.frac = None
            self.genus = None
        self.is_admissible = is_admissible
    
    def __str__(self):
        frac = '' if self.frac == 1 else self.frac
        vert_or_ray = 'Ray' if self._is_ray else 'Vertex' 
        return '{} {}: represented by {}*S_{},{} at {}, mapped from surface with index {}'.format(vert_or_ray, self.index, self.frac, self.genus, self.num_boundary_comps, self.coords, self.surface_index)

    def detail(self):
        return self.__str__()

class Ray(AdornedVertex):
    pass

class DualVertex(Vertex):
    def __init__(self, index, coords, dual_face):
        Vertex.__init__(self, index, coords)
        self.dual_face = dual_face


class Facet():
    def __init__(self, index, dim, vertices, rays):
        self.index = index
        self.dim = dim
        self.vertices = vertices
        self.rays = rays










import sys
import argparse
import tnorm.GUI.application

start = tnorm.GUI.application.start

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='tnorm GUI')
    parser.add_argument('load', nargs='?', help='path to load from when starting')
    args = parser.parse_args()
    
    start(load_from=args.load)


#-*-python-*-
from __future__ import print_function

import regina
import snappy
import sys
import time
import itertools


from tnorm.kernel.euler import *
from tnorm.kernel.boundary import *
from tnorm.kernel.homology import *
from tnorm.kernel.regina_helpers import *
from tnorm.kernel.matrices import *
from tnorm.norm_ball import *
from tnorm.sage_types import *
from tnorm.utilities import *
from tnorm.x3d_to_html import *

import tnorm.constants

QUIET = tnorm.constants.QUIET

#preparser(False)


class TN_wrapper():
	"""
	A class for working with Thurston Norm on H2(M,\partial M).

	sage: W = TN_wrapper(SnapPeaTri)

	SnapPeaTri can be any of the following:
	(1) A snappy Manifold
	(2) a regina.SnapPeaTriangulation triangulation
	(3) Any string argument that snappy.Manifold() can take, e.g.,

	sage: W = TN_wrapper('L6a5')
	sage: W = TN_wrapper('/path/to/snappy/file.tri')
	sage: W = TN_wrapper(string)       # where string is the triangulation encoded as a string (i.e., contents of a tri file)

	"""
	def __init__(self, SnapPeaTri, qtons=None, tracker=False, quiet=False, allows_non_admissible=False, basis='natural'):
		self.H1_basis = basis
		
		if isinstance(SnapPeaTri,str):
			self.manifold = snappy.Manifold(SnapPeaTri)
		elif isinstance(SnapPeaTri, regina.engine.SnapPeaTriangulation):
			self.manifold = snappy.Manifold(SnapPeaTri.snapPea())
		elif isinstance(SnapPeaTri, snappy.Manifold):
			self.manifold = SnapPeaTri
		
		if self.H1_basis == 'shortest':
			self.manifold.set_peripheral_curves('shortest')
		self.triangulation = regina.SnapPeaTriangulation(self.manifold._to_string())
		self._angle_structure = solve_lin_gluingEq(self.triangulation)
		self._intersection_matrices = [intersection_mat(self.manifold, self.triangulation, cusp) for cusp in range(self.triangulation.countCusps())]
		self._qtons = qtons
		self._tkr = False
		self.allows_non_admissible = allows_non_admissible
		QUIET = quiet

		# qtons memoization caches-----
		self._euler_char = {}
		self._bdy_slopes_unoriented = {}
		self._map_to_H1bdy = {}
		self._map_to_ball = {}
		self._map_to_H2 = {}
		self._num_boundary_comps = {}
		self._num_bdy_comps_unoriented = {}
		self._over_face = {}
		self._is_norm_minimizing = {}
		self._is_admissible = {}

		if not QUIET:			
			print('Enumerating quad transversely oriented normal surfaces (qtons)... ', end='')
		if tracker:
			self._tkr = regina.ProgressTracker()
		self._qtons = self.qtons()
		for i in range(self._qtons.size()):
			self._qtons.surface(i).setName(str(i))
		if not QUIET:
			print('Done.')

	def qtons(self):
		"""
		Enumerate vertex oriented quad normal surfaces for the ideal triangulation self.triangulation()
		This is cached in self.qtons, so calls to this function after the initial call (when TN_wrapper() is 
		instantiated) will be fast. 

		Returns a regina.NormalSurfaces object.
		"""
		if self._qtons == None:
			if self._tkr:
				if self.allows_non_admissible:
					return regina.NormalSurfaces.enumerate(self.triangulation,regina.NS_ORIENTED_QUAD,regina.NS_IMMERSED_SINGULAR,regina.NS_ALG_DEFAULT,self._tkr)
				else:
					return regina.NormalSurfaces.enumerate(self.triangulation,regina.NS_ORIENTED_QUAD,regina.NS_VERTEX,regina.NS_ALG_DEFAULT,self._tkr)
			else:
				if self.allows_non_admissible:
					return regina.NormalSurfaces.enumerate(self.triangulation,regina.NS_ORIENTED_QUAD,regina.NS_IMMERSED_SINGULAR,regina.NS_ALG_DEFAULT)
				else:
					return regina.NormalSurfaces.enumerate(self.triangulation,regina.NS_ORIENTED_QUAD,regina.NS_VERTEX,regina.NS_ALG_DEFAULT)
		else:
			return self._qtons


	def show_progress(self):
		print(self._tkr.percent())


	def euler_char(self,qtons):
		"""
		Return the Euler characteristic of the spun normal surface. 
		Argument can be a surface (Regina oriented spun normal surface) or the index of a surface.

		sage: W.euler_char(3)    # return Euler characteristic of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		-1

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())
			
		if ind in self._euler_char:
			return self._euler_char[ind]
		else:
			s = self.qtons().surface(ind)
			ec = euler_char_(s,self._angle_structure)
			self._euler_char[ind] = ec
			return ec


	def bdy_slopes_unoriented(self,qtons):
		"""
		Return the boundary slopes of the spun normal surface. This will be different than the return value
		of Regina's method boundaryIntersections(), because Regina returns the intersection numbers,
		not the slope (i.e., Regina's method will be the negative inverse of these slopes). 
		Argument can be a surface (Regina oriented spun normal surface) or the index of a surface.

		sage: W.boundary_slopes(3)    # return boundary slopes of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		[(-1, 0), (0, 1), (-1, 0)]

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._bdy_slopes_unoriented:
			return self._bdy_slopes_unoriented[ind]
		else:
			s = self.qtons().surface(ind)
			bs = bdy_slopes_unoriented_(s, self)
			self._bdy_slopes_unoriented[ind] = bs
			return bs

	def boundary_slopes(self,qtons):
		"""
		Return the image of the given surface in H1(\partial M). Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		sage: W._map_to_H1bdy(3)    # return image in H1(\partial M) of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		[(-1, 0), (0, 1), (-1, 0)]

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._map_to_H1bdy:
			return self._map_to_H1bdy[ind]
		else:
			s = self.qtons().surface(ind)
			h1 = qtons_to_H1bdy(s, self)
			self._map_to_H1bdy[ind] = h1
			return h1
	
	def map_to_H2(self,qtons):
		"""
		Return the image of the given surface in H2(M, \partial M). Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		sage: W.mapToH2(3)    # return the image in H2(M,\partial M) of the surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		(0,1,0)

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._map_to_H2:
			return self._map_to_H2[ind]
		else:
			s = self.qtons().surface(ind)
			h2 = homology_map(s, self)
			self._map_to_H2[ind] = h2
			return h2

	def over_face(self, qtons, as_string=False):

		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._over_face:
			f = self._over_face[ind]
		else:
			v = self.map_to_H2(ind)
			f = over_face_(v,self.norm_ball.polyhedron)
			self._over_face[ind] = f
		if as_string:
			return None if f==None else '<{}>'.format(' '.join([str(v.index()) for v in f.vertices()]))
		else:
			return f

	def is_norm_minimizing(self, qtons):

		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._is_norm_minimizing:
			return self._is_norm_minimizing[ind]
		else:
			v = self.map_to_H2(ind)
			if not self.norm_ball.polyhedron.interior_contains(v):
				self._is_norm_minimizing[ind] = True
				return True
			else:
				self._is_norm_minimizing[ind] = False
				return False

	def is_admissible(self, qtons):
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._is_admissible:
			return self._is_admissible[ind]
		else:
			s = self.qtons().surface(ind)
			mat = oriented_quads_mat(s)
			for row in mat:
				nonzero = [i for i in range(0,6,2) if ( row[i]!=0 or row[i+1]!=0 )]
				if len(nonzero) > 1:
					self._is_admissible[ind] = False
					return False
			self._is_admissible[ind] = True
			return True


	def map_to_ball(self, qtons):
		"""
		Return the image of the given surface in H2(M, \partial M), then divide by Euler characteristic. Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		"""
		try:
			ind = int(qtons)
		except TypeError:
			ind = int(qtons.name())

		if ind in self._map_to_ball:
			return self._map_to_ball[ind]
		else:
			image_in_H2 = self.map_to_H2(ind)
			if self.euler_char(ind) < 0:
				mtb = tuple([Rational(i)/(-self.euler_char(ind)) for i in image_in_H2])
				self._map_to_ball[ind] = mtb
				return mtb
			else:
				self._map_to_ball[ind] = Infinity
				return Infinity



	def genus(self,qtons):
		"""
		Return the genus. Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		sage: W.genus(3)    # return genus of the surface with index 3, i.e, W.qtons().surface(3).
		1
		"""
		return (2-self.euler_char(qtons)-self.num_boundary_comps(qtons))/2
	
	def num_bdy_comps_unoriented(self,qtons):
		"""
		Return the number of boundary components. Argument can be a surface (regina oriented
		spun normal surface) or the index of a surface.

		sage: W.num_boundary_comps(3)    # return num of boundary components of surface with index 3, i.e, W.qtons().surface(3).
		2

		"""	
		try:
			ind = int(qtons)
			s = self.qtons().surface(ind)
		except TypeError:
			ind = int(qtons.name())
			s = qtons

		if ind in self._num_bdy_comps_unoriented:
			return self._num_bdy_comps_unoriented[ind]
		else:
			nbc = num_boundary_comps_unoriented_(s)
			self._num_bdy_comps_unoriented[ind] = nbc
			return nbc


	def num_boundary_comps(self,qtons):
		"""
		Return the number of boundary components after mapping to H1(bdy m). This will be the number
		of boundary components after cancelling oppositely oriented pairs (i.e., a 3-puntured sphere
		will become a once-punctured torus if two of the punctures are on the cusps and boundary curves
		with opposite orientations). Argument can be a surface (regina oriented spun normal surface) or 
		the index of a surface.

		sage: W.numBoundaryComps(3)    # return num of boundary components of surface with index 3, i.e, W.OrientedNormalSurfaceList.surface(3).
		2

		"""	
		try:
			ind = int(qtons)
			s = self.qtons().surface(ind)
		except TypeError:
			ind = int(qtons.name())
			s = qtons

		if ind in self._num_boundary_comps:
			return self._num_boundary_comps[ind]
		else:
			nbc = 0
			b = self.boundary_slopes(ind)
			for slope in b:
				nbc += gcd(slope[0],slope[1])
			self._num_boundary_comps[ind] = nbc
			return nbc

		
	def _image_in_H2M(self):
		image = []
		for i in range(self.qtons().size()):
			pt = self.map_to_H2(i)
			if not pt.is_zero():
				ec = self.euler_char(i)
				if (pt,ec) not in [tup[1:] for tup in image]:
					image.append((i,pt,ec))
		return image

	@cached_property
	def _norm_ball_points(self):
		"""
		Return the points in H2(M,bdy M) which are images of (normalized) vertices of the oriented spun
		normal surface projective solution space with negative Euler characteristic. The convex hull of 
		these points is the norm ball.
		"""
		points = dict()
		rays = dict()  # if the Euler char of s is 0 we can't normalize. This means the norm ball is
		               # infinite in the direction of map_to_H2(s), so it contains the ray (0,map_to_H2(s)).
		image_w_euler = self._image_in_H2M()
		image_w_euler.sort(key=lambda x: x[0])
		image_w_euler.reverse() # we sort and reverse so that the normal surface associated to this vertex 
		                        # is the one of smallest qtons index, among all qtons that map to it.
		for (i,pt,ec) in image_w_euler:
			if ec < 0:
				points[tuple([coord/(-ec) for coord in pt])] = i # divide by euler char to normalize
			elif ec == 0:
				rays[tuple(pt)] = i
		return points, rays

	@cached_property
	def norm_ball(self):
		"""
		Return the norm ball as a sage Polyhedron. Some things you can do:

		sage: B=W.norm_ball
		sage: B.faces(2)        # list the 2-dimensional faces
		sage: B.faces(0)        # list the 0-dimensional faces (vertices)

		sage: F = B.faces(2)
		sage: F[0].vertices()   #return vertices of the 0 face

		sage: B.plot()          # plot of B. I imagine this only works if B is 1,2, or 3 dimensional.

		The Polyhedron class has a lot of methods, and I have not explored many of them. There is 
		probably a lot more that would be useful. Use tab completion to explore!
		"""
		if not QUIET:
			print('Computing Thurston norm unit ball... ', end='')
		pts_dict, rays_dict = self._norm_ball_points
		if len(rays_dict) != 0:
			polyhedron = Polyhedron(vertices=Matrix(pts_dict.keys()), rays=Matrix(rays_dict.keys()), base_ring=QQ, backend='cdd')
			rays = tuple([(rays_dict[tuple(i*vector(v))],i*vector(v)) for v in polyhedron.lines_list() for i in [1,-1]])
			Rays = [Ray(i,rays[i][0],rays[i][1],self.num_boundary_comps(rays[i][0]), self.euler_char(rays[i][0]), self.boundary_slopes(rays[i][0]), True, self.is_admissible(rays[i][0]),True) for i in range(len(rays))]
			vertices = tuple([(pts_dict[tuple(v)],vector(v)) for v in polyhedron.vertices_list()])
			projected_verts = {}
			Vertices = []
			for i,v in vertices:
				v = vector(v)
				v_proj, coeffs = orthogonal_proj(v, polyhedron.lines_list())
				v_proj = tuple(v_proj)
				boundary_slopes = []
				num_boundary_comps = 0
				euler_char = coeffs[0]*self.euler_char(pts_dict[tuple(v)])
				for j in range(self.manifold.num_cusps()):
					slope = vector((0,0))
					slope += vector(self.boundary_slopes(i)[j])*coeffs[0]
					for k in range(len(polyhedron.lines_list())):
						slope += vector(self.boundary_slopes(rays_dict[tuple(polyhedron.lines_list()[k])])[j])*coeffs[k+1]
					boundary_slopes.append(slope)
					num_boundary_comps += gcd(slope[0],slope[1])
				if v_proj not in projected_verts:
					projected_verts[v_proj] = (num_boundary_comps, euler_char, boundary_slopes)
			polyhedron = Polyhedron(vertices=Matrix(projected_verts.keys()), rays=Matrix(rays_dict.keys()), base_ring=QQ, backend='cdd')
			for i in range(len(polyhedron.vertices_list())):
				v = tuple(polyhedron.vertices_list()[i])
				if v in pts_dict:
					Vertices.append(AdornedVertex(i,pts_dict[v],vector(v),self.num_boundary_comps(pts_dict[v]), self.euler_char(pts_dict[v]), self.boundary_slopes(pts_dict[v]), False, self.is_admissible(pts_dict[v]), True))
				else:
					Vertices.append(AdornedVertex(i, None, vector(v), projected_verts[v][0], projected_verts[v][1], projected_verts[v][2],False, None, True))



		else:
			polyhedron = Polyhedron(vertices=Matrix(pts_dict.keys()), base_ring=QQ)
			vertices = tuple([(pts_dict[tuple(v)],vector(v)) for v in polyhedron.vertices_list()])
			Vertices = [AdornedVertex(i,vertices[i][0],vertices[i][1],self.num_boundary_comps(vertices[i][0]), self.euler_char(vertices[i][0]), self.boundary_slopes(vertices[i][0]), False, self.is_admissible(vertices[i][0]), True) for i in range(len(vertices))]
			Rays = []
		ball = TNormBall(Vertices, Rays, polyhedron)
		if not QUIET:
			print('Done.')
		return ball

	@cached_property
	def dual_norm_ball(self):
		B = self.norm_ball
		P = B.polyhedron
		poly_vertices = []
		vertices = []
		for i in range(len(P.faces(P.dim()-1))):
			face = P.faces(P.dim()-1)[i]
			e = face.as_polyhedron().equations_list()[0]
			v = -vector(e[1:])/(e[0])
			poly_vertices.append(v)
			vertices.append(DualVertex(i, v, B.facets(P.dim()-1)[i]))
		P_dual = Polyhedron(vertices=poly_vertices, base_ring=QQ)
		dual_ball = DualNormBall(vertices, P_dual)
		return dual_ball

	@cached_property
	def qtons_info(self):
		if not QUIET:
			print('Analyzing quad transversely oriented normal surfaces... ', end='')
		qtons_info_dict = dict([(i,{}) for i in range(self.qtons().size())])
		for i in range(self.qtons().size()):
			qtons_info_dict[i]['image_in_H2'] = self.map_to_H2(i)
			qtons_info_dict[i]['euler_char'] = self.euler_char(i)
			qtons_info_dict[i]['boundary_slopes'] = self.boundary_slopes(i)
			qtons_info_dict[i]['num_boundary_comps'] = self.num_boundary_comps(i)
			qtons_info_dict[i]['genus'] = (2-self.euler_char(i)-self.num_boundary_comps(i))/2
			qtons_info_dict[i]['is_norm_minimizing'] = self.is_norm_minimizing(i)
			qtons_info_dict[i]['over_face'] = self.over_face(i,as_string=True)
		if not QUIET:
			print('Done.')
		return qtons_info_dict

def check_subfaces(ray,f_poly):
	if f_poly.dim() == 0:
		return f_poly
	for f in f_poly.faces(f_poly.dim()-1):
		if not f.as_polyhedron().intersection(ray).is_empty():
			return check_subfaces(ray,f.as_polyhedron())
		else:
			continue
	return f_poly
			

def over_face_(v, P):
	if v.is_zero():
		return None
	else:
		ray = Polyhedron(rays=[v]) # ray generated by v
		f_poly = check_subfaces(ray, P) # find the face that v lies over, as a polyhedron

		# f_poly is the face we want, but it is in the form of a polyhedron, not a face of P. So we find the face of P with the same vertices, and return that.
		for i in range(P.dim()):
			for f in P.faces(i):
				if sorted([vert.vector() for vert in f.vertices()]) == sorted([vert.vector() for vert in f_poly.vertices()]):
					return f

def lies_over_face(v, P, face):
	# find the lowest dimensional face that v lies over
	v_face = over_face_(v, P)
	# check if v_face found above is a subface of face
	if set(v_face.vertices()).issubset(set(face.vertices())) and set(v_face.lines()).issubset(set(face.lines())):
		return True
	return False

def basis_over_face(basis, P, face):  # check if all vectors in basis lie over face.
	for vec in basis:
		if not lies_over_face(vec, P, face):
			return False
	return True

def orthogonal_proj(v, rays):
	coeffs = [QQ(1)]
	for w in rays:
		w = vector(w)
		coeff = QQ(-(v.dot_product(w)/w.dot_product(w)))
		v = v + coeff*w
		coeffs.append(coeff)
	l = lcm([coeff.denominator() for coeff in coeffs])
	if l != 1:
		coeffs = [coeff*l for coeff in coeffs]
		v = v*l
	return v, coeffs





















import os
import tempfile
import shutil

from tnorm.sage_types import *


def fix_x3d_str(x3d_str):
    x3d_str = x3d_str.replace('\n','')
    x3d_split = x3d_str.split('>')
    x3d_tags = []
    for string in x3d_split:
        split_str = string.split('<')
        for ss in split_str:
            if ss.replace(' ','') != '':
                x3d_tags.append(ss)
    x3d_str = ''
    for tag in x3d_tags:
        if tag[-1] == '/':
            tag_name = tag.split(' ')[0]
            formatted_tag = '<'+tag[:-1]+'></{}>'.format(tag_name)
        else:
            formatted_tag = '<'+tag+'>'
        x3d_str += formatted_tag
    return x3d_str

def make_x3d_text(text,coords,scale):
    text = "\n<Transform translation='{}' scale='{}'><Billboard axisOfRotation='0 0 0'><Shape><Text string='{}' solid='false'></Text><Appearance><Material diffuseColor='1.0 1.0 1.0' shininess='1.0' specularColor='1.0 1.0 1.0'></Material></Appearance></Shape></Billboard></Transform>".format(coords,scale,text)
    return text

def make_x3d_vertex(coords,radius):
    vert = "\n<Transform translation='{}'><Shape>\n<Sphere ccw='true' lit='true' radius='{}' solid='true' subdivision='24,24'>\n</Sphere>\n<Appearance>\n<Material diffuseColor='1.0 1.0 1.0' shininess='1.0' specularColor='1.0 1.0 1.0'>\n</Material>\n</Appearance>\n</Shape>\n</Transform>".format(coords,radius)
    return vert

def rnd(x):
    if abs(x) < .00001:
        return 0
    else:
        return n(x,20)   

def rnd_vec(v):
    return vector(map(rnd,v))

#def scale_vec(v, factor):
#    return v*factor

def make_x3d_schlegel(schlegel_projection, scale_factor):
    SP = schlegel_projection
    edges = SP.lines
    coords = map(lambda v: v*scale_factor,map(rnd_vec,SP.coordinates_of(SP.points)))
    x3d_str = '\n'
    x3d_str += "<Shape>\n<Appearance><Material ambientIntensity='0' emissiveColor='1 1 1' diffuseColor='1 1 1' specularColor='1 1 1' shininess='0.0078125' transparency='0'></Material></Appearance><IndexedLineSet coordIndex='{}' colorPerVertex='false'>\n<Coordinate point='{}'>\n</Coordinate>\n</IndexedLineSet>\n</Shape>"
    coordIndex = ''
    coordPoint = str(coords).replace('0.00000','0.0').replace('[(','').replace(')]','').replace(',','').replace('(','').replace(')',',')
    for p in SP.lines:
        coordIndex += '{} {} -1, '.format(p[0],p[1])
    x3d_str = x3d_str.format(coordIndex, coordPoint)
    return x3d_str


def make_x3d_html(norm_ball, online=False, directory=None, dual=False):
    if dual:
        title = 'Dual TNorm Ball'
    else:
        title = 'TNorm Ball'
    text = ''
    P=norm_ball.polyhedron

    if P.dim() == 4:
        SP = P.schlegel_projection()
        hull=Polyhedron(SP.coordinates_of(SP.points))
        bounding_box = hull.bounding_box()
        scale_factor = 3/(max(bounding_box[1]))
        x3d_str = make_x3d_schlegel(SP, scale_factor)
        vertices_coords = map(lambda x: n(x,20),SP.coordinates_of(SP.points))

    else:
        bounding_box = P.bounding_box()
        max_coord = max(bounding_box[1])
        scale_factor = 3./max_coord
        x3d_str = P.render_solid().transform(scale=(scale_factor,scale_factor,scale_factor)).x3d_str()
        x3d_str = fix_x3d_str(x3d_str).replace('><','>\n<')
        vertices_coords = P.vertices_list()

    vertices_str = ''
    labels_str = ''
    for i in range(len(vertices_coords)):
        v = vertices_coords[i]
        scaled_v = n(vector(v),20)*scale_factor
        normalized_v = n(vector(v),20).normalized()
        label_coords = scaled_v + normalized_v*.25
        str_vert_coords = '{} {} {}'.format(scaled_v[0], scaled_v[1], scaled_v[2])
        str_label_coords = '{} {} {}'.format(label_coords[0], label_coords[1], label_coords[2])
        v = norm_ball.vertex(i)
        if dual:
            label = '{}'.format(i)
        else:
            frac = '1/{}*'.format(-v.euler_char) if v.euler_char!=-1 else ''
            label = '{}S_{},{}'.format(frac,v.genus,v.num_boundary_comps)
        vertices_str += make_x3d_vertex(str_vert_coords,'.045')
        labels_str += make_x3d_text(label,str_label_coords,'.3 .3 .3')

    xmin, xmax = n(bounding_box[0][0]*scale_factor,20), n(bounding_box[1][0]*scale_factor,20)
    ymin, ymax = n(bounding_box[0][1]*scale_factor,20), n(bounding_box[1][1]*scale_factor,20)
    zmin, zmax = n(bounding_box[0][2]*scale_factor,20), n(bounding_box[1][2]*scale_factor,20)

    datadir = os.path.dirname(__file__)
    path = os.path.join(datadir, 'data', 'x3d_templates', 'axes.html')
    with open(path,'r') as f:
        axis_len = max(xmax-xmin+2, ymax-ymin+2, zmax-zmin+2)
        axes = ''.join(f.readlines()).format(axis_len, axis_len/2., axis_len/2.+.2)


    if online:
        head = "<head>\n<title>{}</title>\n<script type='text/javascript' src='http://www.x3dom.org/x3dom/release/x3dom.js'>\n</script>\n<link rel='stylesheet' type='text/css' href='http://www.x3dom.org/x3dom/release/x3dom.css'>\n</link>\n<h1>{}</h1>\n<style>\n</style>\n</head>\n".format(title,text)
    else:
        head = "<head>\n<title>{}</title>\n<script type='text/javascript' src=x3dom.js>\n</script>\n<link rel='stylesheet' type='text/css' href=x3dom.css></link>\n<h1>{}</h1>\n<style>\n</style>\n</head>\n".format(title,text) 




#    xaxes = "\n<Transform rotation='1 0 0 0' translation='0 0 0'>\n<Shape>\n<Appearance><Material ambientIntensity='0' emissiveColor='0 0 0' diffuseColor='0 0 0' specularColor='0 0 0' shininess='0.0078125' transparency='0'></Material></Appearance><IndexedLineSet coordIndex='0 1 -1 2 1 -1 3 1 -1' colorPerVertex='false'>\n<Coordinate point='{} 0 0, {} 0 0, '>\n</Coordinate>\n</IndexedLineSet>\n</Shape>\n</Transform>".format(xmin-2,xmax+2,)



    body = "<body>\n<x3d width='1000px' height='700px', margin-left='300px'>\n<scene><Background backUrl='[]' bind='false' bottomUrl='[]' crossOrigin='""' description='""' frontUrl='[]' groundAngle='[]' groundColor='(0,0,0)' isActive='false' leftUrl='[]' rightUrl='[]' scaling='[]' skyAngle='[]' skyColor='(0,0,0)' topUrl='[]' transparency='0/1' ></Background>\n{}\n</scene>\n</x3d>\n</body>\n".format(x3d_str+labels_str+vertices_str+axes)

    page = "<html>\n{}{}</html>".format(head,body)

    datadir = os.path.dirname(__file__)
    path = os.path.join(datadir, 'data', 'x3dom')
    css = os.path.join(path, 'x3dom.css')
    js = os.path.join(path, 'x3dom.js')

    shutil.copy2(css, directory)
    shutil.copy2(js, directory)

    filename = os.path.join(directory, 'normball_x3d.html')
    
    with open(filename,'w') as file:
        file.write(page)
    return os.path.abspath(filename)



















from __future__ import print_function

import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

this_directory = os.path.dirname(__file__)
source_directory = os.path.join(this_directory, 'tnorm')
exec(open(os.path.join(source_directory, 'version.py')).read())  # Load in the variable __version__.


setuptools.setup(
    name='tnorm',  
    version=__version__,
    author="William Worden",
    author_email="wtworden@gmail.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wtworden/TNorm",
    packages=["tnorm"],
    package_data={
        'tnorm.GUI': ['images/spinpoly/*'],
        },
    classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
    ],
 )